#!/usr/bin/perl
# Copyright (C) 2017 Yuri V. Kravatsky, jiri@eimb.ru - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the Creative Commons CC BY-NC-SA 3.0 license.
# If you use this script in your scienfic projects please cite:
# Kravatsky YV, Chechetkin VR, Fedoseeva DM, Gorbacheva MA, Kravatskaya GI, Kretova OV, Tchurikov NA.
# A bioinformatic pipeline for monitoring of the mutational stability of viral drug targets with deep-sequencing technology.
# DOI: XX.XXXX/XXXX, PMID: XXXXXX

use Bio::SeqIO;
use Bio::SearchIO;
use File::Copy;

### This is FASTA parameters. Tune it according to your needs. In the most cases you could want to change
### only E-value that corresponds to alignment quality. Look to FASTA manual for further details
    my $options = "-3 -E 0.001 -q -n";
### Change it to the path of your RAMdrive or the most fast HDD. RAMdrive is strongly recommended
### Under Linux RAMdrive support is built-in and can be activated by adding to /etc/fstab the 
### following string that creates 256M RAMdrive
### tmpfs /mnt/ramdrive tmpfs rw,size=256M     0       0
### then issue command "sudo mount -a" to mount RAMdrive without reboot
    my $RAMDRIVE="/mnt/ramdrive/";
### The path to your Smith-Waterman search utility from FASTA Package. 
    my $FASTASRC="./ssearch35";
### ATTENTION! Fit this variable to the length of your reference sequence. If you don't expect 
### non-complete reads or significant amount of deletions mutations you should set this value 
### to the exact length of your reference sequence or +/- 1bp from that length
    my $TARGETLENGTH=30;
### ATTENTION! This is the string for the FIRST run to check/tune the most invariable target inside YOUR data!
### You should make the first run, then look to output file, choose the most invariable sequence, set it 
### as reference sequence, change $FIRSTRUN to 0 and re-start the whole procedure.
    my $FIRSTRUN=1;  # comment this string after you've set the most invariable sequence as a reference one
#    my $FIRSTRUN=0; # UNcomment this string after you've set the most invariable sequence as a reference one

    my $ARGC=scalar @ARGV;

    if ($ARGC!=3) { die "Usage: find_invariable aligned.fastafile targetfile outfile\n"; }

    $aligned=$ARGV[0];
    $targetfile=$ARGV[1];
    $outputfile=$ARGV[2];

    open (TARG,"$targetfile") || die "Can't read from \"$targetfile\": $!";
    open (OUTP,">$outputfile") || die "Can't write to \"$outputfile\": $!";

    my $target;

    while(<TARG>) {
       $target = $target . $_;
                  }
    close (TARG);
    print "$target";

    my $seqio = Bio::SeqIO->new( -file => $aligned,
                                 -format => "fasta",
                                 -alphabet  => "dna" );
    my %targets;
    my $totaltargets=0;


# Copy to RAMdrive ssearch and target to speedup FASTA runs
    my $fasta   = $RAMDRIVE . "ssearch36";
    if (copy ($FASTASRC, $fasta)==0) { die("Cannot copy fasta executable to RAMdrive: $!\n"); }
    chmod 0555, $fasta;
    my $library = $RAMDRIVE . $aligned;
    my $ftarg = $RAMDRIVE . $targetfile;
    if (copy ($targetfile, $ftarg)==0) { die("Cannot copy target file to RAMdrive: $!\n"); }

nextrecord: while (my $fastaseq = $seqio->next_seq) {
#write temporary file with FASTA database
                   my $seq_out = Bio::SeqIO->new(-file   => ">$library",
                                                 -format => "fasta",
                                                 -alpabet => "dna" );
                   $seq_out->write_seq($fastaseq);
# create command line to run FASTA search
                   my $command = "$fasta $options $ftarg $library";
# run FASTA search, capture output
                   open $fh,"$command |" || die("Cannot run fasta cmd of $command: $!\n");
                   my $in  = Bio::SearchIO->new(-format => 'fasta', -fh => $fh);
                   while( my $result = $in->next_result ) {
                       while( my $hit = $result->next_hit ) {
## $hit is a Bio::Search::Hit::HitI compliant object
                           while( my $hsp = $hit->next_hsp ) {
## $hsp is a Bio::Search::HSP::HSPI compliant object
                               $query_str = $hsp->seq_str('query');
                               $hit_str = $hsp->seq_str('hit');

                               my ($query_beg, $query_end) = $hsp->range('query');
                               my ($hit_beg, $hit_end) = $hsp->range('hit');
                               $query_str =~ s/^( *)//g;
                               my $offset=length($1);
                               my $leng=length($query_str);
                               my $final=substr $hit_str, $offset, $leng;
                               $final =~ s/^( *)//g;
                               if (length($final)<$TARGETLENGTH) { goto nextrecord; }
                               $targets { $final } ++;
                               $totaltargets++;
                               if ($FIRSTRUN) { if ($totaltargets>=1000) { 
                                                       print "ATTENTION!!!! It's first evaluation run. Check the first string in the file \"$outputfile\" and transfer it into target FASTA file!\nChange FIRSTRUN variable value to 1 in the string 31 of this script file,\nthen repeat processing from the very beginning (2align.sh)!\n";
                                                       goto fin;         } 
                                              }
                               goto nextrecord;              }
                                                            }
                                                           }
                                            }
fin:    print "$totaltargets targets processed\n";

        my $totalhits=0;
        foreach my $targvalue (sort { $targets{$b} <=> $targets{$a} } keys %targets) {
            $totalhits += $targets{$targvalue};
            my $perc=100.0 * $targets{$targvalue} / $totaltargets;
            my $c=sprintf "%s\t%d\t%.3f\n", $targvalue, $targets{$targvalue}, $perc;  
            print OUTP $c;                                                           }
        print OUTP "Total $totalhits hits found\n";
        close (OUTP);
