#!/usr/bin/perl
# Copyright (C) 2017 Yuri V. Kravatsky, jiri@eimb.ru - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the Creative Commons CC BY-NC-SA license.
# If you use this script in your scienfic projects please cite:
# Kravatsky YV, Chechetkin VR, Fedoseeva DM, Gorbacheva MA, Kravatskaya GI, Kretova OV, Tchurikov NA.
# A bioinformatic pipeline for monitoring of the mutational stability of viral drug targets with deep-sequencing technology.
# DOI: XX.XXXX/XXXX, PMID: XXXXXX

     use Statistics::Descriptive;
     use Bio::AlignIO;

    my $ARGC=scalar @ARGV;

    if ($ARGC==0||$ARGC>3) { die "Usage: mut_stat.pl clustalfile.aln \n"; }

    my $infile=$ARGV[0];

    my $in = Bio::AlignIO->new(-file   => $ARGV[0],
                               -format => "clustalw" );
      
    my $basename=substr($infile,0,rindex($infile,"."));
    my $outfile=$basename.".stat";
    open (OUTP,">$outfile") || die "Can't create \"$outfile\": $!";

    my $numseq;
    while ( my $aln = $in->next_aln() ) {
       if ($aln->is_flush!=1) { print STDERR "Warning: Alignment is not box, check output thouroughly!\n"; }

       $numseq=$aln->num_sequences;
       my $alignlength=$aln->length;
       my $matchline = $aln->match_line();

#print OUTP "Length: $alignlength\n";
#print OUTP "Number of sequences: $numseq\n";

       my @seqarr;
       my $i=0;
       foreach $seq ($aln->each_seq) {
#           $seqarr[$i] = $seq->subseq(1, $alignlength);
            $seqarr[$i] = $seq;
           $i++;                     }

#for (my $i=0;$i<$numseq;$i++) { 
#     my $currseq=$seqarr[$i]->subseq(1, $alignlength);
#     print OUTP "$currseq\n"; 
#                              }

#print OUTP "$matchline\n";

       my @matcharr=split('',$matchline);
       my $matchleng=scalar @matcharr;
#print OUTP "Alignment length is $alignlength, match array length is $matchleng\n";

       my @transitions;
       my @transversions;
       my $realnumseq=$numseq-1;
print "Really real number of sequences is $realnumseq\n";

       for (my $i=0;$i<$realnumseq;$i++) {
          for (my $j=0;$j<3;$j++) {
               $transitions[$i][$j]=0;
	       $transversions[$i][$j]=0;
       	                          }      }

#### here we should detect effective length (cut ------ at the end)
#       my $reallength=$alignlength;
#       for (my $i=$alignlength;$i>0;$i--) {
#          my $emptychar=0;
#          for (my $j=0;$j<$numseq;$j++) { 
#               if ($seqarr[$j]->subseq($i,$i) eq "-") { $emptychar++; }
#                                        }
#          if ($emptychar<$numseq) { last; } 
#          $reallength=$i;  
#                                          }

#### Cut all "-" at the end - align to block
       my $reallength=$alignlength;
#cuttail: for (my $i=$alignlength;$i>0;$i--) {
#          for (my $j=0;$j<$numseq;$j++) { 
#               if ($seqarr[$j]->subseq($i,$i) eq "-") { next cuttail; }
#                                        }
#          $reallength=$i;  
#          last;                             }

#print OUTP "Detected length=$alignlength, Real length=$reallength\n";

#convert all string to upper character
       for (my $j=0;$j<$numseq;$j++) {
          my $seq=$seqarr[$j]->seq;
          $seqarr[$j]->seq(uc($seq)); 
                                     }

       my (@delarr,  @insarr);
       my (@AT, @AG, @AC, @TA, @TG, @TC, @GA, @GT, @GC, @CA, @CT, @CG);

       for (my $i=0;$i<$reallength;$i++) {
            $AT[$i]=$AG[$i]=$AC[$i]=$TA[$i]=$TG[$i]=$TC[$i]=$GA[$i]=$GT[$i]=$GC[$i]=$CA[$i]=$CT[$i]=$CG[$i]=0;
                                         }

       for (my $i=0;$i<$reallength;$i++) {
              for (my $j=0;$j<$realnumseq;$j++) {
                   $delarr[$j][$i]=0;
                   $insarr[$j][$i]=0;           }
              if ($matcharr[$i] eq "*" ) { next; }
              my $numsym=$i+1;
              $wildsym=$seqarr[$realnumseq]->subseq($numsym,$numsym);
              for (my $j=0;$j<$realnumseq;$j++) {
                    my $currsym=$seqarr[$j]->subseq($numsym, $numsym);
                    if ($currsym eq $wildsym) { next; }
                    $pos=$i%3;
# my $numseq=$j+1;
# $numpos=$pos+1;
# print OUTP "$numpos:$numsym:$numseq $wildsym => $currsym\n";
# Insertions: Reference "-" => any
                    if (($wildsym eq "-") && ($currsym ne "-")) { $insarr[$j][$i]++;
                                                                  next; }
# Deletions:  Reference any => "-"
                    if (($wildsym ne "-") && ($currsym eq "-")) { $delarr[$j][$i]++; 
                                                                  next; }
                    if ($wildsym eq "C") {
                         if ($currsym eq "T") { $transitions[$j][$pos]++;
                                                $CT[$i]++;
                                                next; }
                         if ($currsym eq "A") { $transversions[$j][$pos]++;
                                                $CA[$i]++;
                                                next; }
                         if ($currsym eq "G") { $transversions[$j][$pos]++;
                                                $CG[$i]++;
                                                next; }
                                         }
                    if ($wildsym eq "A") {
                         if ($currsym eq "G") { $transitions[$j][$pos]++;
                                                $AG[$i]++;
                                                next; }
                         if ($currsym eq "C") { $transversions[$j][$pos]++;
                                                $AC[$i]++;
                                                next; }
                         if ($currsym eq "T") { $transversions[$j][$pos]++;
                                                $AT[$i]++;
                                                next; }
                                         }
                    if ($wildsym eq "G") {
                         if ($currsym eq "A") { $transitions[$j][$pos]++; 
                                                $GA[$i]++;
                                                next; }
                         if ($currsym eq "T") { $transversions[$j][$pos]++; 
                                                $GT[$i]++;
                                                next; }
                         if ($currsym eq "C") { $transversions[$j][$pos]++; 
                                                $GC[$i]++;
                                                next; }
                                         }
                    if ($wildsym eq "T") {
                         if ($currsym eq "C") { $transitions[$j][$pos]++; 
                                                $TC[$i]++;
                                                next; }
                         if ($currsym eq "A") { $transversions[$j][$pos]++; 
                                                $TA[$i]++;
                                                next; }
                         if ($currsym eq "G") { $transversions[$j][$pos]++; 
                                                $TG[$i]++;
                                                next; }
                                         }
                    print STDERR "Impossible situation, unknown mutation!\n";
                                                }
                                          }

       my @insertions;
       my @deletions;

       for (my $j=0;$j<$realnumseq;$j++) {
          my @currarr;
          for (my $i=0;$i<$reallength;$i++) {
             $currarr[$i]=$insarr[$j][$i];  }
          push (@insertions, indels(\@currarr,$reallength));
          for (my $i=0;$i<$reallength;$i++) {
             $currarr[$i]=$delarr[$j][$i];  }
          push (@deletions, indels(\@currarr,$reallength));
                                         }

       my @transitsum;
       my @transversum;
       my @transitfreq;
       my @transverfreq;
       my @totalsum;
       my @totalfreq;

       my @transitseq;
       my @transverseq;
       my @sumseq;
       for (my $i=0;$i<$realnumseq;$i++) {
           $transitseq[$i]=0;
           $transverseq[$i]=0;
           for (my $j=0;$j<3;$j++) {
             $transitseq[$i]+=$transitions[$i][$j];
             $transverseq[$i]+=$transversions[$i][$j];
                                   }     
           $sumseq[$i]=$transitseq[$i]+$transverseq[$i]; }

       for (my $j=0;$j<3;$j++) {
          $transitsum[$j]=0;
          $transversum[$j]=0;
          for (my $i=0;$i<$realnumseq;$i++) {
              $transitsum[$j]+=$transitions[$i][$j];
              $transversum[$j]+=$transversions[$i][$j];
                                            }
          $transitfreq[$j]=$transitsum[$j]/($reallength*$realnumseq);
          $transverfreq[$j]=$transversum[$j]/($reallength*$realnumseq);
          $totalsum[$j]=$transitsum[$j]+$transversum[$j];
          $totalfreq[$j]=$totalsum[$j]/($reallength*$realnumseq);
                               } 
       my $threshold=3.84/$realnumseq;
       print OUTP "Sequences/Total\t$realnumseq\t$reallength\n";
#       print OUTP "Sum by pos\nTotal:\t$totalsum[0]\t$totalsum[1]\t$totalsum[2]\n";
#       print OUTP "Transitions:\t$transitsum[0]\t$transitsum[1]\t$transitsum[2]\n";
#       print OUTP "Transversions:\t$transversum[0]\t$transversum[1]\t$transversum[2]\n";

#       my $c=sprintf("Freq by pos\nTotal:\t%.6f\t%.6f\t%.6f\n",$totalfreq[0],$totalfreq[1],$totalfreq[2]);
#       print OUTP $c;
#       $c=sprintf("Transitions:\t%.6f\t%.6f\t%.6f\nTransversions:\t%.6f\t%.6f\t%.6f\n",$transitfreq[0],$transitfreq[1],$transitfreq[2],$transverfreq[0],$transverfreq[1],$transverfreq[2]);
#       print OUTP $c;

       my $globalsum=$totalsum[0]+$totalsum[1]+$totalsum[2];
       my $globalfreq=$globalsum/($realnumseq*$reallength);
       $c=sprintf("Mutations Total/Frequency\t$globalsum\t%.5f\n",$globalfreq);
       print OUTP $c;


       my $totalins=scalar @insertions;
       my $totaldel=scalar @deletions;
       my @y=();

       my $statinsert = Statistics::Descriptive::Full->new();
       $statinsert->add_data(@y);

       $statinsert->add_data(@insertions);
       my $median=$statinsert->median();
       my $stddev=$statinsert->standard_deviation();
       my $mean=$statinsert->mean();
       my $min=$statinsert->min();
       my $max=$statinsert->max();

#print OUTP "Insertions\t@insertions\n";

       my $freqins=$totalins/($reallength*$realnumseq);
       $c=sprintf("Insertions\t$mean\t$median\t$stddev\t$max\t$min\t$totalins\t%.8f\n",$freqins);
       print OUTP $c;
       undef $statinsert;

       my $statdelete = Statistics::Descriptive::Full->new();
       $statdelete->add_data(@y);

       $statdelete->add_data(@deletions);
       $median=$statdelete->median();
       $stddev=$statdelete->standard_deviation();
       $mean=$statdelete->mean();
       $min=$statdelete->min();
       $max=$statdelete->max();

#print OUTP "Deletions\t@deletions\n";
       my $freqdel=$totaldel/($reallength*$realnumseq);
       $c=sprintf("Deletions\t$mean\t$median\t$stddev\t$max\t$min\t$totaldel\t%.8f\n",$freqdel);
       print OUTP $c;
       undef $statdelete;
##       print OUTP "Transitions/seq";
##       for (my $i=0;$i<$realnumseq;$i++) { print OUTP "\t$transitseq[$i]"; }
##       print OUTP "\nTransversions/seq";
##       for (my $i=0;$i<$realnumseq;$i++) { print OUTP "\t$transverseq[$i]"; }
##       print OUTP "\nMutations/seq";
##       for (my $i=0;$i<$realnumseq;$i++) { print OUTP "\t$sumseq[$i]"; }
##       print OUTP "\n";

       print OUTP "\nMutations per nucleotide position\n";
       print OUTP "\t\tAT\tAG\tAC\tTA\tTG\tTC\tGA\tGT\tGC\tCA\tCT\tCG\tsum\tthreshold\n";

       my @summut;
       for (my $i=0;$i<$reallength;$i++) {
             my $pos=$i+1;
             my $sum=($AT[$i]+$AG[$i]+$AC[$i]+$TA[$i]+$TG[$i]+$TC[$i]+$GA[$i]+$GT[$i]+$GC[$i]+$CA[$i]+$CT[$i]+$CG[$i])/$realnumseq;
             $summut[$i]=$sum;
             my $currsym=$seqarr[$realnumseq]->subseq($pos,$pos);

             $AT[$i]/=$realnumseq; 
             $AG[$i]/=$realnumseq; 
             $AC[$i]/=$realnumseq; 

             $TA[$i]/=$realnumseq; 
             $TG[$i]/=$realnumseq; 
             $TC[$i]/=$realnumseq; 

             $GA[$i]/=$realnumseq; 
             $GT[$i]/=$realnumseq; 
             $GC[$i]/=$realnumseq; 

             $CA[$i]/=$realnumseq; 
             $CT[$i]/=$realnumseq; 
             $CG[$i]/=$realnumseq; 

             if ($AT[$i]==0) { $AT[$i]=''; }
             if ($AG[$i]==0) { $AG[$i]=''; }
             if ($AC[$i]==0) { $AC[$i]=''; }

             if ($TA[$i]==0) { $TA[$i]=''; }
             if ($TG[$i]==0) { $TG[$i]=''; }
             if ($TC[$i]==0) { $TC[$i]=''; }

             if ($GA[$i]==0) { $GA[$i]=''; }
             if ($GT[$i]==0) { $GT[$i]=''; }
             if ($GC[$i]==0) { $GC[$i]=''; }

             if ($CA[$i]==0) { $CA[$i]=''; }
             if ($CT[$i]==0) { $CT[$i]=''; }
             if ($CG[$i]==0) { $CG[$i]=''; }

             print OUTP "$pos\t$currsym\t$AT[$i]\t$AG[$i]\t$AC[$i]\t$TA[$i]\t$TG[$i]\t$TC[$i]\t$GA[$i]\t$GT[$i]\t$GC[$i]\t$CA[$i]\t$CT[$i]\t$CG[$i]\t$sum\t$threshold\n";
                                         }

       my @y=();
       my $stat = Statistics::Descriptive::Full->new();
       $stat->add_data(@y);
       $stat->add_data(@summut);
       my $median=$stat->median();
       my $mean=$stat->mean();
       my $stddev=$stat->standard_deviation();
       my $minimum=$stat->min();
       my $maximum=$stat->max();
       my $lower=$stat->quantile(1);
       my $upper=$stat->quantile(3);
       my $low5=$stat->percentile(5);
       my $up95=$stat->percentile(95);
       my $c = sprintf "%.7f\t%.7f\t%.7f\t%.7f\t%.7f\t[%.7f, %.7f]\t[%.7f, %.7f]\n",$mean,$median,$stddev,$minimum,$maximum,$low5,$up95,$lower,$upper;
       print OUTP "Mean\tMedian\tStdDev\tMin\tMax\t[5%, 95%]\t[25%, 75%]\n$c";
       undef $stat;
                                       }

sub indels {
    my $arr=$_[0];
    my $leng=$_[1];
    my @groups;

    for (my $i=0;$i<$leng;$i++) {
      if ($arr->[$i]==0) { next; }
      my $sum=0;
      while ($arr->[$i]>0) {
        if ($arr->[$i]>1) { print STDERR "Error at index $i, number of indels is $arr->[$i] !!! \n"; }
        $sum++;
        $i++;              }
      push (@groups, $sum);     }
    return (@groups);
}
